<?php
/**
 * Ágora 3.0
 *
 * Un marco de desarrollo web sencillo.

  Copyright (c) 2009-2025 Guillermo Martínez J.

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.

 */

/**
 * Identifica la versión de Ágora.
 */
  define ('AG_VERSION', '3.0');

/**
 * Directorio donde se encuentra ágora.
 */
  if (!defined ('AG_DIR'))
    define ('AG_DIR', __DIR__ . DIRECTORY_SEPARATOR);

/**
 * Directorio de vistas.
 */
  if (!defined ('AG_DIR_VISTAS'))
    define ('AG_DIR_VISTAS', AG_DIR . 'vistas' . DIRECTORY_SEPARATOR);

/**
 * Nombre de la vista para errores 404.
 */
  if (!defined ('AG_VISTA_404'))
    define ('AG_VISTA_404', 'error404');

/**
 * Nombre de la vista para errores genéricos.
 */
  if (!defined ('AG_VISTA_ERROR'))
    define ('AG_VISTA_ERROR', 'error');

/**
 * Directorio de módulos.
 */
  if (!defined ('AG_DIR_MODULOS'))
    define ('AG_DIR_MODULOS', AG_DIR . 'modulos' . DIRECTORY_SEPARATOR);

/**
 * Caducidad por defecto de las huellas (cookies) en segundos.
 */
  if (!defined ('AG_CADUCIDAD_HUELLAS'))
    define ('AG_CADUCIDAD_HUELLAS', 12 * 60 * 60);



/**
 * Clase contenedora de Ágora.
 *
 * Puede usarse como base para otras clases, creando un objeto, o llamando
 * directamente a los métodos estáticos.
 */
  class Agora
  {
  /**
   * Devuelve el valor de la variable de configuración, si existe, o el valor
   * $Defecto si no existe.
   *
   * La configuración se almacena en un diccionario global llamado $config.
   * Este diccionario no es creado por Ágora.
   *
   * Este método busca, primero, con el nombre del valor tal cual.  Si no
   * encuentra el valor, vuelve a buscarlo convirtiendo el valor a minúsculas.
   * Si sigue sin encontrarlo entonces devuelve el valor por defecto o "false"
   * si este no se indicó.
   * @param string $aNombre Nombre del valor de configuración.
   * @param mixed  $aValDefecto Valor por defecto.
   * @return mixed Devuelve el valor de configuración o $aValDefecto, o false.
   */
    public static function config (string $aNombre, mixed $aDefecto = false)
      : mixed
    {
      global $config;

      if (isset ($config[$aNombre])) return $config[$aNombre];
      $aNombre = strtolower ($aNombre);
      return isset ($config[$aNombre]) ? $config[$aNombre]: $aDefecto;
    }



  /**
   * Cambia los caracteres '&', '%', '<' y '>' para evitar inyecciones.
   * Es posible que no evite absolutamente todas las inyecciones, pero he hecho
   * algunas pruebas y funciona razonablemente bien.
   *
   * Otra utilidad que tiene es que si se le pasa código fuente (incluso XML),
   * lo devuelve en un formato que permite mostrarlo en el navegador sin temor
   * a que ejecute código.
   * @param mixed $aTexto Texto a filtrar, o un array con los textos a procesar.
   * @return mixed Texto filtrado o un array con el contenido filtrado.
   */
    public static function filtraXSS (mixed $aTexto): mixed
    {
    # Si es una lista, procesa cada elemento.
      if (is_array ($aTexto)):
	foreach ($aTexto as $ndx => $valor)
          $aTexto[$ndx] = str_replace (
            array ('&', '%', '<', '>'),
            array ('&amp;', '&#37;', '&lt;', '&gt;'),
            $aTexto
          );
	return $aTexto;
      endif;
    # Realiza el filtrado.
      if (is_string ($aTexto))
        return str_replace (
          array ('&', '%', '<', '>'),
          array ('&amp;', '&#37;', '&lt;', '&gt;'),
          $aTexto
        );
      else
	return $aTexto;
    }



  /**
   * Función de ayuda para obtener elementos de una lista.
   *
   * Este método busca, primero, con el nombre del valor tal cual.
   * Si no encuentra el valor, vuelve a buscarlo convirtiendo el nombre a
   * minúsculas.
   * @param array $aLista La lista conteniendo los elementos.
   * @param string $aNombre Nombre del elemento.
   * @param bool $aFiltraXSS Indica si debe o no filtrar el elemento.
   * @return mixed El elemento o false si no lo encuentra.
   */
    public static function tomaElementoLista (
      array $aLista,
      ?string $aNombre=null,
      bool $aFiltraXSS=false
    ): mixed
    {
    # Si no solicita un elemento, devuelve la lista completa.
      if ($aNombre === null):
        if (!empty ($aLista))
          return ($aFiltraXSS)
            ? self::filtraXSS ($aLista)
            : $aLista;
        else
          return false;
      endif;
    # Si no encuentra el elemento, tal vez exista en "minúsculas".
      if (!isset ($aLista[$aNombre])) $aNombre = strtolower ($aNombre);
    # Devuelve el elemento solicitado, si existe.
      if (isset ($aLista[$aNombre]))
	return ($aFiltraXSS)
	     ? self::filtraXSS ($aLista[$aNombre])
	     : $aLista[$aNombre];
    # No existe.
      return false;
    }



  /**
   * Devuelve el valor del parámetro GET solicitado o .F. si este no se
   * recibió.
   * @param string $aNombre Nombre del valor.
   * @param bool $aFiltraXSS Indica si debe o no filtrar el valor.
   * @return string El valor o false si no lo encuentra.
   */
    public static function get (
      ?string $aNombre = null,
      bool $aFiltraXSS = false
    ): string
    {
      return self::tomaElementoLista ($_GET, $aNombre, $aFiltraXSS);
    }



  /**
   * Devuelve el valor del parámetro POST solicitado o .F. si este no se
   * recibió.
   * @param string $aNombre Nombre del valor.
   * @param bool $aFiltraXSS Indica si debe o no filtrar el valor.
   * @return string El valor o false si no lo encuentra
   */
    public static function post (
      ?string $aNombre = null,
      bool $aFiltraXSS = false
    ): string
    {
      return self::tomaElementoLista ($_POST, $aNombre, $aFiltraXSS);
    }



  /**
   * Devuelve el valor del parámetro POST o GET solicitado o .F. si este no se
   * recibió.
   *
   * Si está en ambas entradas, POST tiene preferencia.
   * @param string $aNombre Nombre del valor.
   * @param bool $aFiltraXSS Indica si debe o no filtrar el valor.
   * @return string El valor o false si no lo encuentra
   */
    public static function get_post (
      string $aNombre,
      bool $aFiltraXSS = false
    ): string
    {
      if (isset ($_POST[$aNombre]) or isset ($_POST[strtolower ($aNombre)]))
	return self::tomaElementoLista ($_POST, $aNombre, $aFiltraXSS);
      return self::tomaElementoLista ( $_GET, $aNombre, $aFiltraXSS);
    }



    private static $_cfgHuellas = array (
      'ruta' => '', 'dominio' => '', 'https' => false, 'javascript' => false,
      'caducidad' => AG_CADUCIDAD_HUELLAS
    );



  /**
   * Modifica la configuración del sistema de huellas (/cookies/).
   * Puede recibir el nombre del parámetro y el valor, donde el parámetro puede
   * ser:
   * * 'caducidad':  Tiempo de duración por defecto, en segundos.
   * * 'dominio': Indica el dominio para las huellas.
   * * 'ruta':  Indica la ruta (/path/) para las huellas.
   * * 'https': Indica si debe usarse el protocolo HTTPS.
   * * 'javascript': Indica si se podrá o no acceder a los valores a través de
   *                 JavaScript.
   * También puede recibir un diccionario (/array/) en el que el índice es el
   * nombre del parámetro.
   *
   * Si el parámetro indicado no existe, símplemente se ignora su valor.
   * @param mixed $aParametro Parámetro a modificar.  Una cadena o un diccionario.
   * @param mixed $aValor Valor a asignar al parámetro.
   */
    public static function ponParametroHuellas (
      mixed $aParametro,
      mixed $aValor=''
    ): void
    {
      if (is_array ($aParametro)):
	foreach ($aParametro as $parametro => $valor):
          $parametro = strtolower ($parametro);
          if (isset (self::$_cfgHuellas[$parametro]))
            self::$_cfgHuellas[$parametro] = $valor;
        endforeach;
      else:
        $aParametro = strtolower ($aParametro);
	if (isset (self::$_cfgHuellas[$aParametro]))
	  self::$_cfgHuellas[$aParametro] = $aValor;
      endif;
    }



  /**
   * Devuelve la configuración de las huellas.
   * Si no se indica el parámetro, devuelve todos en forma de diccionario.
   * @param string $aParametro Nombre del parámetro.
   * @return mixed Valor del parámetro solicitado o diccionario con todos ellos.
   */
    public static function tomaParametroHuellas (?string $aParametro=null)
      : mixed
    {
      return $aParametro === null
        ? self::$_cfgHuellas
        : self::$_cfgHuellas[strtolower ($aParametro)];
    }



  /**
   * Establece el valor de una huella.
   * @param string $aNombre Nombre de la huella.  Es independiente de las mayúsculas.
   * @param string $aValor Valor a asignar a la huella.
   * @param int $aCaducidad Tiempo de vida de la huella, en segundos.
   * @return bool .F. en caso de que no se pudiera establecer el valor.
   */
    public static function ponHuella (
      string $aNombre, string $aValor, ?int $aCaducidad=null
    ): bool
    {
      if ($aCaducidad === null) $aCaducidad = self::$_cfgHuellas['caducidad'];
      if ($aCaducidad > 0) $aCaducidad += time ();
      return setcookie (strtolower ($aNombre), $aValor, $aCaducidad,
	self::$_cfgHuellas['ruta'],
	self::$_cfgHuellas['dominio'],
	self::$_cfgHuellas['https'],
	!self::$_cfgHuellas['javascript']
      );
    }



  /**
   * Devuelve el valor de una huella o .F. si esta no se encontró.
   * @param string $aNombre Nombre de la huella.  Es independiente de las mayúsculas.
   * @param bool $aFiltraXSS Indica si el contenido debe filtrarse.
   * @return mixed El valor de la huella o .F. si no lo encontró.
   */
    public static function tomaHuella (string $aNombre, bool $aFiltraXSS = false)
      : string
    {
      return self::tomaElementoLista ($_COOKIE, $aNombre, $aFiltraXSS);
    }



  /**
   * Elimina la huella indicada, de existir.
   * @param string $aNombre Nombre de la huella.  Es independiente de las mayúsculas.
   */
    public static function eliminaHuella (string $aNombre): void
    {
      self::ponHuella ($aNombre, '', time () - 3600);
    }



  /**
   * Elimina todas las huellas disponibles.
   */
    public static function limpiaHuellas (): void
    {
      foreach (array_keys ($_COOKIE) as $nombre) self::eliminaHuella ($nombre);
    }



  /**
   * Carga y procesa la vista.
   *
   * @param string $aVista Nombre de la vista, sin extensión.
   * @param array $aVars Diccionario con las variables usadas por la vista.
   * @param bool $aDevolver Devolver la vista procesada como cadena (true) o enviar el resultado al cliente (false).
   */
    public static function cargaVista (
      string $aVista,
      array $aVars = array (),
      bool $aDevolver = false
    ): string
    {
      $lNombreArchivo =  AG_DIR_VISTAS . "$aVista.php";
      if (!is_file ($lNombreArchivo))
	throw new \Exception ("No se encontró la plantilla '$aVista'", 404);
    # Si recibes una lista de parámetros, genera las variables precisas.
    # Nótese que no sobreescribe las ya existentes para impedir la modificación
    # de las variables globales.
      if (count ($aVars) > 0) extract ($aVars, EXTR_SKIP);
    # Genera la vista.
      ob_start ();
      if ((bool) @ini_get ('short_open_tag') === false):
      # El intérprete no permite etiqueta de apertura corta ("<?="), así que
      # sustitúyelo por "<?php echo ", que es equivalente.  Esto lo he tomado
      # prestado de CodeIgniter 2.
	echo eval (
	  '?'.'>' . preg_replace (
	    '/;*\s*\?'.'>/',
	    '; ?'.'>',
	    str_replace (
	      '<'.'?=',
	      '<'.'?php echo ',
	      file_get_contents ($lNombreArchivo)
	    )
	  )
	);
      else:
	require ($lNombreArchivo);
      endif;
    # Comprueba si debes devolver el resultado.
      if ($aDevolver):
	$lResultado = ob_get_contents ();
	ob_end_clean ();
	return $lResultado;
      endif;
      ob_end_flush ();
      return ''; # Siempre tiene que devolver algo.
    }



  /**
   * Contenedor de módulos.
   */
    private static TModulos $_modulos;



  /**
   * Carga el archivo del módulo indicado y crea un objeto.
   *
   * Recibe el nombre del módulo y, opcionalmente, el nombre del objeto.
   *
   * Observe que el nombre del módulo se convertirá a minúsculas, mientras que
   * el de la clase debe estar en MAYÚSCULAS ó CamelCase, así, si el módulo se
   * llama "Modulo", el archivo se llamará "modulo.php" y la clase "MODULO" ó
   * "Modulo".
   *
   * Si se carga el módulo varias veces con el mismo nombre de objeto, ignora
   * las cargas adicionales.  Nótese que los nombres "Modulo", "modulo",
   * "MODULO", etc. hacen referencia al mismo módulo.
   *
   * Si como nombre de objeto se pasa un valor "false" entonces no crea ningún
   * objeto.
   *
   * En caso de haber creado un objeto a partir de la clase Agora, se puede
   * cargar el módulo directamente desde la propiedad "modulo". Por ejemplo:
   *
   * $agora = new \Agora\Agora ();
   * $agora->modulos->carga ('MiModulo');
   * $agora->modulos->miModulo->hazAlgo ();
   */
    public static function cargaModulo (
      string $aNombreModulo,
      mixed $aNombreObjeto = ''
    ): mixed
    {
      if (!isset (self::$_modulos)) self::$_modulos = new TModulos ();
      return self::$_modulos->carga ($aNombreModulo, $aNombreObjeto);
    }



  /**
   * Devuelve un objeto de módulo.
   *
   * En caso de haber creado un objeto a partir de la clase Agora, se puede
   * acceder a los objetos de módulo como propiedades del objeto. Por ejemplo:
   *
   * $agora = new \Agora\Agora ();
   * $agora->cargaModulo ('MiModulo');
   * $agora->miModulo->hazAlgo ();
   *
   * Otra alternativa:
   *
   * $agora->modulos->miModulo->hazAlgo ();
   *
   * En caso de que el objeto no exista, se devolverá null.
   *
   * El nombre del objeto no depende de las mayúsculas.
   */
     public static function objetoModulo (string $aNombre): mixed
     {
       if (isset (self::$_modulos->$aNombre)) return self::$_modulos->$aNombre;
       return null;
     }



  /**
   * Lista de códigos de error HTTP.
   */
    public static $errorHTTP = array (
	200 => 'OK',
	201 => 'Created',
	202 => 'Accepted',
	203 => 'Non-Authoritative Information',
	204 => 'No Content',
	205 => 'Reset Content',
	206 => 'Partial Content',

	300 => 'Multiple Choices',
	301 => 'Moved Permanently',
	302 => 'Found',
	304 => 'Not Modified',
	305 => 'Use Proxy',
	307 => 'Temporary Redirect',

	400 => 'Bad Request',
	401 => 'Unauthorized',
	403 => 'Forbidden',
	404 => 'Not Found',
	405 => 'Method Not Allowed',
	406 => 'Not Acceptable',
	407 => 'Proxy Authentication Required',
	408 => 'Request Timeout',
	409 => 'Conflict',
	410 => 'Gone',
	411 => 'Length Required',
	412 => 'Precondition Failed',
	413 => 'Request Entity Too Large',
	414 => 'Request-URI Too Long',
	415 => 'Unsupported Media Type',
	416 => 'Requested Range Not Satisfiable',
	417 => 'Expectation Failed',

	500 => 'Internal Server Error',
	501 => 'Not Implemented',
	502 => 'Bad Gateway',
	503 => 'Service Unavailable',
	504 => 'Gateway Timeout',
	505 => 'HTTP Version Not Supported'
      );



  /**
   * Envía al cliente una cabecera HTTP.
   * @param int $aCodigo El código de la cabecera.
   * @param string $aTexto Texto de la cabecera.
   * @return El "$aTexto" definitivo.
   */
    public static function enviaCabeceraHTTP (
      int $aCodigo = 500,
      string $aTexto = ''
    ): string
    {
    # Obtén el mensaje de cabecera, según su código y asignación.
      if ($aCodigo == 0) $aCodigo = 500;
      if ($aTexto == ''):
	if (isset (self::$errorHTTP[$aCodigo])):
	  $aTexto = self::$errorHTTP[$aCodigo];
	else:
	  self::enviaError (
	    "No se encontró cabecera para el error '$aCodigo'.&nbsp; Por favor, use el código adecuado o provea su propia cabecera."
	  );
	endif;
      endif;
    # Envía la cabecera.
      $protocolo = (isset ($_SERVER['SERVER_PROTOCOL']))
		 ? $_SERVER['SERVER_PROTOCOL']
		 : false;
      if (substr (php_sapi_name (), 0, 3) == 'cgi')
	header ("Status: $aCodigo $aTexto", true);
      elseif ($protocolo == 'HTTP/1.1' or $protocolo == 'HTTP/1.0')
	header ("$protocolo $aCodigo $aTexto", true, $aCodigo);
      else
	header ("HTTP/1.1 $aCodigo $aTexto", true, $aCodigo);
      return $aTexto;
    }



  /**
   * Redirige al cliente a otra URL.
   */
    public static function Redirige (string $aURL, int $aRazon = 307): void
    {
    # TODO: Comprobar que la URL sea válida?
      self::enviaCabeceraHTTP ($aRazon);
      header ("Location: $aURL");
    }



  /**
   * Devuelve al cliente una página de error y termina.
   *
   * La página de error está indicada por la constante AG_VISTA_ERROR, y usa las
   * variables siguientes:
   * * $codigo: Número del código de error.
   * * $nombre: Texto que define el error HTTP.
   * * $mensaje: Mensaje de error.
   * Este método termina la ejecución del programa.
   * @param mixed $aMensaje El mensaje de error a mostrar.  Texto o array.
   * @param int $aCodigo El código del error HTTP.
   * @param string $aNombre Nombre del error.
   */
    public static function enviaError (
      mixed $aMensaje,
      int $aCodigo = 500,
      string $aNombre = ''
    ): void
    {
    # Envía la cabecera.
      if ($aCodigo == 0) $aCodigo = 500;
      $aNombre = self::enviaCabeceraHTTP ($aCodigo, $aNombre);
    # Envía la página de error.
      if (is_array ($aMensaje)) $aMensaje = implode ('</p><p>', $aMensaje);
      self::cargaVista (AG_VISTA_ERROR, array (
	'nombre' => $aNombre,
	'codigo' => $aCodigo,
	'mensaje' => "<p>$aMensaje</p>"
      ));
    # Termina la ejecución.
      exit;
    }



  /**
   * Igual que "enviaError", solo que envía un error 404 y usa la plantilla
   * indicada por la constante AG_VISTA_404, que usa las variables "$url" y
   * "$mensaje".
   * @param string $aURL La URL de la página.
   * @param mixed $aMensaje El mensaje de error a mostrar.  Texto o array.
   */
    public static function enviaError404 (
      string $aURL,
      mixed $aMensaje = 'Página no encontrada.'
    ): void
    {
    # Envía la cabecera.
      self::enviaCabeceraHTTP (404);
    # Envía la página de error.
      if (is_array ($aMensaje)) $aMensaje = implode ('</p><p>', $aMensaje);
      self::CargaVista (AG_VISTA_404, array (
	'url' => $aURL,
	'mensaje' => "<p>$aMensaje</p>"
      ));
    # Termina la ejecución.
      exit;
    }



  /**
   * Ayuda para mostrar un error de "Propiedad no encontrada".
   *
   * Este método debe llamarse desde el método mágico "__get".
   * @param string $aNombre Nombre de la propiedad.  Puede (y debe) incluir la clase y el espacio de nombre.
   */
    public static function muestraErrorFaltaPropiedad (string $aNombre): void
    {
    # Obtén el trazado.
      $trazado = debug_backtrace (
        DEBUG_BACKTRACE_PROVIDE_OBJECT|DEBUG_BACKTRACE_IGNORE_ARGS
      );
    # Obtén el punto donde se intentó acceder a la propiedad.
      $ndx = 1;
      while ($trazado[$ndx+1]['function'] == '__get') $ndx++;
    # Lanza el error.
      trigger_error (
        "Undefined property <code>$aNombre</code> in <b>{$trazado[$ndx]['file']}</b> on line <b>{$trazado[$ndx]['line']}</b><br />Error handler",
        E_USER_WARNING
      );
    }


  /**
   * Constructor.
   */
    public function __construct ()
    {
      self::$_modulos = new TModulos ();
    }



  /**
   * Acceso a los objetos de módulo como propiedad.
   */
    public function __get (string $aNombre): mixed
    {
      $nombrePropiedad = strtolower ($aNombre);
      if ($nombrePropiedad == 'modulos') return self::$_modulos;
      if (isset (self::$_modulos->$nombrePropiedad))
        return self::$_modulos->$nombrePropiedad;
    # No lo encontró, así que trata de replicar el error de PHP.
      $this->muestraErrorFaltaPropiedad ('Agora::$'.$aNombre);
    # Tiene que devolver algo.
      return false;
    }



  /**
   * Comprueba si el objeto fue creado.
   *
   * Ojo:  si intenta comprobarse un objeto usando la propiedad "modulos" (o sea,
   * así: isset ($agora->modulos->miModulo)) el resultado puede ser erróneo.
   */
    public function __isset (string $aNombre): bool
    {
      if (strtolower ($aNombre) == 'modulos') return true;
      return isset (self::$_modulos->$aNombre);
    }
  }




/**
 * Para gestionar módulos y objetos de módulo.
 *
 * Es usada internamente por Ágora.
 */
  class TModulos
  {
    private array $_objetos = array ();
    private array $_modulos = array ();



  /**
   * Acceso a los objetos de módulo.
   */
    public function __get (string $aNombre): mixed
    {
      $nombrePropiedad = strtolower ($aNombre);
      if (isset ($this->_objetos[$nombrePropiedad]))
        return $this->_objetos[$nombrePropiedad];
    # Didn't found it so try to replicate PHP's error.
      $trazado = Agora::obtenPuntoErrorTrazado ('__get');
      trigger_error (
        "Undefined property \$$aNombre in <b>{$trazado['file']}</b> on line <b>{$trazado['line']}</b><br />error handler",
        E_USER_WARNING
      );
      return false;
    }



  /**
   * Comprueba si el objeto fue creado.
   */
    public function __isset (string $aNombre): bool
    {
      return isset ($this->_objetos[strtolower ($aNombre)]);
    }



  /**
   * Respuesta a la asignación de propiedades.
   *
   * Este método existe porque el mensaje de error de PHP es confuso y no ayuda
   * en absoluto.
   */
    public function __set (string $aNombre, mixed $aValor): void
    {
      $trazado = debug_backtrace (
        !DEBUG_BACKTRACE_PROVIDE_OBJECT|DEBUG_BACKTRACE_IGNORE_ARGS,
        2
      );
      $infoError = isset ($trazado[1]) ? $trazado[1] : $trazado[0];
      trigger_error (sprintf (
        'Undefined property $%s in <b>%s</b> on line <b>%d</b><br />',
        $aNombre,
        $infoError['file'],
        $infoError['line']
      ), E_USER_WARNING);
    }



  /**
   * Permite destruir un objeto.
   */
    public function __unset (string $aNombre): void
    {
      $nombrePropiedad = strtolower ($aNombre);
        if (isset ($this->_objetos[$nombrePropiedad]))
          unset ($this->_objetos[$nombrePropiedad]);
    }



  /**
   * Carga el archivo del módulo indicado y crea un objeto.
   *
   * Recibe el nombre del módulo y, opcionalmente, el nombre del objeto.
   *
   * Observe que el nombre del módulo se convertirá a minúsculas, mientras que
   * el de la clase debe estar en MAYÚSCULAS ó CamelCase, así, si el módulo se
   * llama "Modulo", el archivo se llamará "modulo.php" y la clase "MODULO" ó
   * "Modulo".
   *
   * Si se carga el módulo varias veces con el mismo nombre de objeto, ignora
   * las cargas adicionales.  Nótese que los nombres "Modulo", "modulo",
   * "MODULO", etc. hacen referencia al mismo módulo.
   *
   * Si como nombre de objeto se pasa un valor "false" entonces no crea ningún
   * objeto.
   */
    public function carga (
      string $aNombreModulo,
      mixed $aNombreObjeto = ''
    ): mixed
    {
    # Nombres e identificadores.
      $idModulo = $aNombreModulo;
      $rutaModulo = '';
      if (strpos ($aNombreModulo, '/') !== false):
        $listaDir = explode ('/', $aNombreModulo);
        $aNombreModulo = end ($listaDir);
        unset ($listaDir[count ($listaDir) - 1]);
        $rutaModulo = implode (DIRECTORY_SEPARATOR, $listaDir).DIRECTORY_SEPARATOR;
      endif;
      if ($aNombreObjeto !== false):
        if ($aNombreObjeto === '')
          $aNombreObjeto = strtolower ($aNombreModulo);
        else
          $aNombreObjeto = strtolower ($aNombreObjeto);
      endif;
    # Comprueba si cargó el módulo.
      if (isset ($this->_modulos[$idModulo])):
      # Obtén el nombre de la clase.
        $nombreClase = $this->_modulos[$idModulo];
      else:
        $nombreArchivo = AG_DIR_MODULOS . "$rutaModulo$aNombreModulo.php";
        if (!is_file ($nombreArchivo)):
          $nombreArchivo = AG_DIR_MODULOS.$rutaModulo.strtolower ($aNombreModulo).'.php';
          if (!is_file ($nombreArchivo))
            throw new \Exception (
              "No pude encontrar el módulo \"$aNombreModulo\"",
              404
            );
        endif;
        include_once $nombreArchivo;
      # Comprueba si hay clase.
        $nombreClase = ucfirst ($aNombreModulo);
        if (!class_exists ($nombreClase)) $nombreClase = '.';
        $this->_modulos[$idModulo] = $nombreClase;
      endif;
    # Crea objeto, si se tercia.
      if ($aNombreObjeto !== false):
      # Comprueba si el objeto existe.
        if (!isset ($this->_objetos[$aNombreObjeto])):
          if ($nombreClase === '.')
            throw new \Exception (
              "No se encontró ninguna clase de módulo en el módulo \"$aNombreModulo\".",
              500
            );
          $this->_objetos[$aNombreObjeto] = new $nombreClase ();
        endif;
        return $this->_objetos[$aNombreObjeto];
      else:
        return null;
      endif;
    }



  /**
   * Permite añadir un objeto a la lista.
   * @param mixed $aObjeto El objeto a insertar.
   * @param string $aNombre El nombre del objeto.  Si ya hay un objeto con dicho nombre, se reemplazará.
   */
    public function insertaObjeto (mixed $aObjeto, string $aNombre)
    {
      $aNombre = strtolower ($aNombre);
      if ($aObjeto = nil)
        unset ($this->_objetos[$nombrePropiedad]);
      else
        $this->_objetos[$nombrePropiedad] = $aObjeto;
    }
  }
